use std::{collections::HashMap, path::PathBuf};

use anyhow::{bail, Context, Result};
use image::io::Reader;

use crate::{handle::TextureHandle, Content};

pub(crate) mod syntax {
	use std::path::PathBuf;

	use serde::Deserialize;
	// Raw serde syntax structs.
	// These are never seen by code outside this crate.

	#[derive(Debug, Deserialize)]
	pub struct Texture {
		pub path: PathBuf,
	}
}

/// Represents a texture that may be used in the game.
#[derive(Debug, Clone)]
pub struct Texture {
	/// The name of this texture
	pub name: String,

	/// The handle for this texture
	pub handle: TextureHandle,

	/// The path to this texture's image file
	pub path: PathBuf,
}

impl crate::Build for Texture {
	type InputSyntax = HashMap<String, syntax::Texture>;

	fn build(texture: Self::InputSyntax, ct: &mut Content) -> Result<()> {
		for (texture_name, t) in texture {
			let path = ct.texture_root.join(t.path);
			let reader = Reader::open(&path).with_context(|| {
				format!(
					"Failed to read texture `{}` from file `{}`",
					texture_name,
					path.display()
				)
			})?;
			let dim = reader.into_dimensions().with_context(|| {
				format!(
					"Failed to dimensions of texture `{}` from file `{}`",
					texture_name,
					path.display()
				)
			})?;

			let h = TextureHandle {
				index: ct.textures.len(),
				aspect: dim.0 as f32 / dim.1 as f32,
			};

			if texture_name == ct.starfield_texture_name {
				if ct.starfield_handle.is_none() {
					ct.starfield_handle = Some(h)
				} else {
					// This can't happen, since this is a hashmap.
					unreachable!("Found two starfield textures! Something is very wrong.")
				}
			}

			ct.texture_index.insert(texture_name.clone(), h);
			ct.textures.push(Self {
				name: texture_name,
				path,
				handle: h,
			});
		}

		if ct.starfield_handle.is_none() {
			bail!(
				"Could not find a starfield texture (name: `{}`)",
				ct.starfield_texture_name
			)
		}

		return Ok(());
	}
}
